import multiprocessing
import sys
import threading
import time

import pytest

import env
from pybind11_tests import gil_scoped as m


class ExtendedVirtClass(m.VirtClass):
    def virtual_func(self):
        pass

    def pure_virtual_func(self):
        pass


def test_callback_py_obj():
    m.test_callback_py_obj(lambda: None)


def test_callback_std_func():
    m.test_callback_std_func(lambda: None)


def test_callback_virtual_func():
    extended = ExtendedVirtClass()
    m.test_callback_virtual_func(extended)


def test_callback_pure_virtual_func():
    extended = ExtendedVirtClass()
    m.test_callback_pure_virtual_func(extended)


def test_cross_module_gil_released():
    """Makes sure that the GIL can be acquired by another module from a GIL-released state."""
    m.test_cross_module_gil_released()  # Should not raise a SIGSEGV


def test_cross_module_gil_acquired():
    """Makes sure that the GIL can be acquired by another module from a GIL-acquired state."""
    m.test_cross_module_gil_acquired()  # Should not raise a SIGSEGV


def test_cross_module_gil_inner_custom_released():
    """Makes sure that the GIL can be acquired/released by another module
    from a GIL-released state using custom locking logic."""
    m.test_cross_module_gil_inner_custom_released()


def test_cross_module_gil_inner_custom_acquired():
    """Makes sure that the GIL can be acquired/acquired by another module
    from a GIL-acquired state using custom locking logic."""
    m.test_cross_module_gil_inner_custom_acquired()


def test_cross_module_gil_inner_pybind11_released():
    """Makes sure that the GIL can be acquired/released by another module
    from a GIL-released state using pybind11 locking logic."""
    m.test_cross_module_gil_inner_pybind11_released()


def test_cross_module_gil_inner_pybind11_acquired():
    """Makes sure that the GIL can be acquired/acquired by another module
    from a GIL-acquired state using pybind11 locking logic."""
    m.test_cross_module_gil_inner_pybind11_acquired()


def test_cross_module_gil_nested_custom_released():
    """Makes sure that the GIL can be nested acquired/released by another module
    from a GIL-released state using custom locking logic."""
    m.test_cross_module_gil_nested_custom_released()


def test_cross_module_gil_nested_custom_acquired():
    """Makes sure that the GIL can be nested acquired/acquired by another module
    from a GIL-acquired state using custom locking logic."""
    m.test_cross_module_gil_nested_custom_acquired()


def test_cross_module_gil_nested_pybind11_released():
    """Makes sure that the GIL can be nested acquired/released by another module
    from a GIL-released state using pybind11 locking logic."""
    m.test_cross_module_gil_nested_pybind11_released()


def test_cross_module_gil_nested_pybind11_acquired():
    """Makes sure that the GIL can be nested acquired/acquired by another module
    from a GIL-acquired state using pybind11 locking logic."""
    m.test_cross_module_gil_nested_pybind11_acquired()


def test_release_acquire():
    assert m.test_release_acquire(0xAB) == "171"


def test_nested_acquire():
    assert m.test_nested_acquire(0xAB) == "171"


def test_multi_acquire_release_cross_module():
    for bits in range(16 * 8):
        internals_ids = m.test_multi_acquire_release_cross_module(bits)
        assert len(internals_ids) == 2 if bits % 8 else 1


# Intentionally putting human review in the loop here, to guard against accidents.
VARS_BEFORE_ALL_BASIC_TESTS = dict(vars())  # Make a copy of the dict (critical).
ALL_BASIC_TESTS = (
    test_callback_py_obj,
    test_callback_std_func,
    test_callback_virtual_func,
    test_callback_pure_virtual_func,
    test_cross_module_gil_released,
    test_cross_module_gil_acquired,
    test_cross_module_gil_inner_custom_released,
    test_cross_module_gil_inner_custom_acquired,
    test_cross_module_gil_inner_pybind11_released,
    test_cross_module_gil_inner_pybind11_acquired,
    test_cross_module_gil_nested_custom_released,
    test_cross_module_gil_nested_custom_acquired,
    test_cross_module_gil_nested_pybind11_released,
    test_cross_module_gil_nested_pybind11_acquired,
    test_release_acquire,
    test_nested_acquire,
    test_multi_acquire_release_cross_module,
)


def test_all_basic_tests_completeness():
    num_found = 0
    for key, value in VARS_BEFORE_ALL_BASIC_TESTS.items():
        if not key.startswith("test_"):
            continue
        assert value in ALL_BASIC_TESTS
        num_found += 1
    assert len(ALL_BASIC_TESTS) == num_found


def _intentional_deadlock():
    m.intentional_deadlock()


ALL_BASIC_TESTS_PLUS_INTENTIONAL_DEADLOCK = ALL_BASIC_TESTS + (_intentional_deadlock,)


def _run_in_process(target, *args, **kwargs):
    if len(args) == 0:
        test_fn = target
    else:
        test_fn = args[0]
    # Do not need to wait much, 10s should be more than enough.
    timeout = 0.1 if test_fn is _intentional_deadlock else 10
    process = multiprocessing.Process(target=target, args=args, kwargs=kwargs)
    process.daemon = True
    try:
        t_start = time.time()
        process.start()
        if timeout >= 100:  # For debugging.
            print(
                "\nprocess.pid STARTED", process.pid, (sys.argv, target, args, kwargs)
            )
            print(f"COPY-PASTE-THIS: gdb {sys.argv[0]} -p {process.pid}", flush=True)
        process.join(timeout=timeout)
        if timeout >= 100:
            print("\nprocess.pid JOINED", process.pid, flush=True)
        t_delta = time.time() - t_start
        if process.exitcode == 66 and m.defined_THREAD_SANITIZER:  # Issue #2754
            # WOULD-BE-NICE-TO-HAVE: Check that the message below is actually in the output.
            # Maybe this could work:
            # https://gist.github.com/alexeygrigorev/01ce847f2e721b513b42ea4a6c96905e
            pytest.skip(
                "ThreadSanitizer: starting new threads after multi-threaded fork is not supported."
            )
        elif test_fn is _intentional_deadlock:
            assert process.exitcode is None
            return 0
        elif process.exitcode is None:
            assert t_delta > 0.9 * timeout
            msg = "DEADLOCK, most likely, exactly what this test is meant to detect."
            if env.PYPY and env.WIN:
                pytest.skip(msg)
            raise RuntimeError(msg)
        return process.exitcode
    finally:
        if process.is_alive():
            process.terminate()


def _run_in_threads(test_fn, num_threads, parallel):
    threads = []
    for _ in range(num_threads):
        thread = threading.Thread(target=test_fn)
        thread.daemon = True
        thread.start()
        if parallel:
            threads.append(thread)
        else:
            thread.join()
    for thread in threads:
        thread.join()


# TODO: FIXME, sometimes returns -11 (segfault) instead of 0 on macOS Python 3.9
@pytest.mark.parametrize("test_fn", ALL_BASIC_TESTS_PLUS_INTENTIONAL_DEADLOCK)
def test_run_in_process_one_thread(test_fn):
    """Makes sure there is no GIL deadlock when running in a thread.

    It runs in a separate process to be able to stop and assert if it deadlocks.
    """
    assert _run_in_process(_run_in_threads, test_fn, num_threads=1, parallel=False) == 0


# TODO: FIXME on macOS Python 3.9
@pytest.mark.parametrize("test_fn", ALL_BASIC_TESTS_PLUS_INTENTIONAL_DEADLOCK)
def test_run_in_process_multiple_threads_parallel(test_fn):
    """Makes sure there is no GIL deadlock when running in a thread multiple times in parallel.

    It runs in a separate process to be able to stop and assert if it deadlocks.
    """
    assert _run_in_process(_run_in_threads, test_fn, num_threads=8, parallel=True) == 0


# TODO: FIXME on macOS Python 3.9
@pytest.mark.parametrize("test_fn", ALL_BASIC_TESTS_PLUS_INTENTIONAL_DEADLOCK)
def test_run_in_process_multiple_threads_sequential(test_fn):
    """Makes sure there is no GIL deadlock when running in a thread multiple times sequentially.

    It runs in a separate process to be able to stop and assert if it deadlocks.
    """
    assert _run_in_process(_run_in_threads, test_fn, num_threads=8, parallel=False) == 0


# TODO: FIXME on macOS Python 3.9
@pytest.mark.parametrize("test_fn", ALL_BASIC_TESTS_PLUS_INTENTIONAL_DEADLOCK)
def test_run_in_process_direct(test_fn):
    """Makes sure there is no GIL deadlock when using processes.

    This test is for completion, but it was never an issue.
    """
    assert _run_in_process(test_fn) == 0



# SPDX-FileCopyrightText: Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cvcuda
import pytest as t
from random import randint


@t.mark.parametrize(
    "num_images, format, min_size, max_size, erasing_area_num, random, seed",
    [
        (1, cvcuda.Format.U8, (100, 100), (200, 200), 1, False, 0),
        (5, cvcuda.Format.RGB8, (100, 100), (200, 100), 1, True, 1),
    ],
)
def test_op_erase(
    num_images, format, min_size, max_size, erasing_area_num, random, seed
):

    parameter_shape = [erasing_area_num]
    anchor = cvcuda.Tensor(parameter_shape, cvcuda.Type._2S32, "N")
    erasing = cvcuda.Tensor(parameter_shape, cvcuda.Type._3S32, "N")
    imgIdx = cvcuda.Tensor(parameter_shape, cvcuda.Type.S32, "N")
    values = cvcuda.Tensor(parameter_shape, cvcuda.Type.F32, "N")

    input = cvcuda.ImageBatchVarShape(num_images)
    output = cvcuda.ImageBatchVarShape(num_images)
    for i in range(num_images):
        w = randint(min_size[0], max_size[0])
        h = randint(min_size[1], max_size[1])
        img_in = cvcuda.Image([w, h], format)
        input.pushback(img_in)
        img_out = cvcuda.Image([w, h], format)
        output.pushback(img_out)

    tmp = cvcuda.erase(input, anchor, erasing, values, imgIdx)
    assert tmp.uniqueformat is not None
    assert tmp.uniqueformat == output.uniqueformat
    for res, ref in zip(tmp, output):
        assert res.size == ref.size
        assert res.format == ref.format

    tmp = cvcuda.erase_into(
        output, input, anchor, erasing, values, imgIdx, random=random, seed=seed
    )
    assert tmp is output

    stream = cvcuda.Stream()
    tmp = cvcuda.erase(
        src=input,
        anchor=anchor,
        erasing=erasing,
        values=values,
        imgIdx=imgIdx,
        random=random,
        seed=seed,
        stream=stream,
    )
    assert tmp.uniqueformat is not None
    assert tmp.uniqueformat == output.uniqueformat
    for res, ref in zip(tmp, output):
        assert res.size == ref.size
        assert res.format == ref.format

    tmp = cvcuda.erase_into(
        src=input,
        dst=output,
        anchor=anchor,
        erasing=erasing,
        values=values,
        imgIdx=imgIdx,
        random=random,
        seed=seed,
        stream=stream,
    )
    assert tmp is output



# SPDX-FileCopyrightText: Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cvcuda
import pytest as t
import numpy as np
import util

RNG = np.random.default_rng(0)


@t.mark.parametrize(
    "input,out_shape,interp,fmt",
    [
        (
            cvcuda.Tensor([5, 16, 23, 4], np.uint8, "NHWC"),
            [5, 132, 15, 4],
            cvcuda.Interp.LINEAR,
            cvcuda.Format.RGB8,
        ),
        (
            cvcuda.Tensor([5, 31, 31, 4], np.uint8, "NHWC"),
            [5, 55, 55, 4],
            cvcuda.Interp.LINEAR,
            cvcuda.Format.RGB8,
        ),
        (
            cvcuda.Tensor([5, 55, 55, 4], np.uint8, "NHWC"),
            [5, 31, 31, 4],
            cvcuda.Interp.LINEAR,
            cvcuda.Format.RGB8,
        ),
        (
            cvcuda.Tensor([5, 16, 23, 4], np.float32, "NHWC"),
            [5, 132, 15, 4],
            cvcuda.Interp.LINEAR,
            cvcuda.Format.RGBf32,
        ),
        (
            cvcuda.Tensor([5, 31, 31, 4], np.float32, "NHWC"),
            [5, 55, 55, 4],
            cvcuda.Interp.LINEAR,
            cvcuda.Format.RGBf32,
        ),
        (
            cvcuda.Tensor([5, 55, 55, 4], np.float32, "NHWC"),
            [5, 31, 31, 4],
            cvcuda.Interp.LINEAR,
            cvcuda.Format.RGBf32,
        ),
    ],
)
def test_op_pillowresize(input, out_shape, interp, fmt):

    out = cvcuda.pillowresize(input, out_shape, fmt, interp)
    assert out.layout == input.layout
    assert out.shape == out_shape
    assert out.dtype == input.dtype

    out = cvcuda.Tensor(out_shape, input.dtype, input.layout)
    tmp = cvcuda.pillowresize_into(out, input, fmt, interp)
    assert tmp is out
    assert out.layout == input.layout
    assert out.shape == out_shape
    assert out.dtype == input.dtype

    stream = cvcuda.Stream()
    tmp = cvcuda.pillowresize_into(
        src=input,
        dst=out,
        format=fmt,
        interp=interp,
        stream=stream,
    )
    assert tmp is out
    assert out.layout == input.layout
    assert out.shape == out_shape
    assert out.dtype == input.dtype


@t.mark.parametrize(
    "nimages, format, max_size, max_pixel, interp",
    [
        (
            5,
            cvcuda.Format.RGB8,
            (16, 23),
            256.0,
            cvcuda.Interp.LINEAR,
        ),
        (
            4,
            cvcuda.Format.RGB8,
            (14, 14),
            256.0,
            cvcuda.Interp.LINEAR,
        ),
        (
            7,
            cvcuda.Format.RGBf32,
            (10, 15),
            256.0,
            cvcuda.Interp.LINEAR,
        ),
    ],
)
def test_op_pillowresizevarshape(
    nimages,
    format,
    max_size,
    max_pixel,
    interp,
):

    input = util.create_image_batch(
        nimages, format, max_size=max_size, max_random=max_pixel, rng=RNG
    )

    base_output = util.create_image_batch(
        nimages, format, max_size=max_size, max_random=max_pixel, rng=RNG
    )

    sizes = []
    for image in base_output:
        sizes.append([image.width, image.height])

    out = cvcuda.pillowresize(input, sizes)
    assert len(out) == len(input)
    assert out.capacity == input.capacity
    assert out.uniqueformat == input.uniqueformat
    assert out.maxsize == base_output.maxsize

    out = cvcuda.pillowresize(
        input,
        sizes,
        interp,
    )
    assert len(out) == len(input)
    assert out.capacity == input.capacity
    assert out.uniqueformat == input.uniqueformat
    assert out.maxsize == base_output.maxsize

    stream = cvcuda.Stream()

    tmp = cvcuda.pillowresize_into(
        src=input,
        dst=base_output,
        interp=interp,
        stream=stream,
    )
    assert tmp is base_output
    assert len(base_output) == len(input)
    assert base_output.capacity == input.capacity
    assert base_output.uniqueformat == input.uniqueformat
    assert base_output.maxsize == base_output.maxsize



# SPDX-FileCopyrightText: Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cvcuda
import pytest as t
import numpy as np


@t.mark.parametrize(
    "input,out_shape,out_layout",
    [
        (cvcuda.Tensor([5, 16, 23, 4], np.uint8, "NHWC"), [5, 4, 16, 23], "NCHW"),
        (cvcuda.Tensor([5, 16, 23, 3], np.uint8, "NHWC"), [5, 3, 16, 23], "NCHW"),
        (cvcuda.Tensor([5, 3, 16, 23], np.uint8, "NCHW"), [5, 16, 23, 3], "NHWC"),
    ],
)
def test_op_reformat(input, out_shape, out_layout):
    out = cvcuda.reformat(input, out_layout)
    assert out.layout == out_layout
    assert out.shape == out_shape
    assert out.dtype == input.dtype

    out = cvcuda.Tensor(out_shape, input.dtype, out_layout)
    tmp = cvcuda.reformat_into(out, input)
    assert tmp is out
    assert out.layout == out_layout
    assert out.shape == out_shape
    assert out.dtype == input.dtype

    stream = cvcuda.Stream()
    out = cvcuda.reformat(src=input, layout=out_layout, stream=stream)
    assert out.layout == out_layout
    assert out.shape == out_shape
    assert out.dtype == input.dtype

    tmp = cvcuda.reformat_into(src=input, dst=out, stream=stream)
    assert tmp is out
    assert out.layout == out_layout
    assert out.shape == out_shape
    assert out.dtype == input.dtype



# SPDX-FileCopyrightText: Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cvcuda
import pytest as t
import numpy as np


@t.mark.parametrize(
    "format,num_images,min_size,max_size,border,bvalue,out_shape,out_layout,out_dtype",
    [
        (
            cvcuda.Format.RGBA8,
            1,
            (10, 5),
            (10, 5),
            cvcuda.Border.REPLICATE,
            0,
            [1, 5, 10, 4],
            "NHWC",
            np.uint8,
        ),
    ],
)
def test_op_padandstack(
    format,
    num_images,
    min_size,
    max_size,
    border,
    bvalue,
    out_shape,
    out_layout,
    out_dtype,
):

    input = cvcuda.ImageBatchVarShape(num_images)

    input.pushback(
        [
            cvcuda.Image(
                (
                    min_size[0] + (max_size[0] - min_size[0]) * i // num_images,
                    min_size[1] + (max_size[1] - min_size[1]) * i // num_images,
                ),
                format,
            )
            for i in range(num_images)
        ]
    )

    left = cvcuda.Tensor([1, 1, num_images, 1], np.int32, "NHWC")
    top = cvcuda.Tensor([1, 1, num_images, 1], np.int32, "NHWC")

    out = cvcuda.padandstack(input, top, left)
    assert out.layout == out_layout
    assert out.shape == out_shape
    assert out.dtype == out_dtype

    out = cvcuda.Tensor(out_shape, out_dtype, out_layout)
    tmp = cvcuda.padandstack_into(out, input, top, left)
    assert tmp is out

    stream = cvcuda.Stream()
    out = cvcuda.padandstack(
        src=input, left=left, top=top, border=border, bvalue=bvalue, stream=stream
    )
    assert out.layout == out_layout
    assert out.shape == out_shape
    assert out.dtype == out_dtype

    tmp = cvcuda.padandstack_into(
        src=input,
        dst=out,
        left=left,
        top=top,
        border=border,
        bvalue=bvalue,
        stream=stream,
    )
    assert tmp is out



# SPDX-FileCopyrightText: Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cvcuda
import pytest as t
import numpy as np
import util

RNG = np.random.default_rng(0)


@t.mark.parametrize(
    "foreground, background, fgMask, outChannels",
    [
        (
            cvcuda.Tensor([5, 9, 9, 3], np.uint8, "NHWC"),
            cvcuda.Tensor([5, 9, 9, 3], np.uint8, "NHWC"),
            cvcuda.Tensor([5, 9, 9, 1], np.uint8, "NHWC"),
            3,
        ),
        (
            cvcuda.Tensor([9, 9, 3], np.uint8, "HWC"),
            cvcuda.Tensor([9, 9, 3], np.uint8, "HWC"),
            cvcuda.Tensor([9, 9, 1], np.uint8, "HWC"),
            4,
        ),
        (
            cvcuda.Tensor([5, 21, 10, 3], np.uint8, "NHWC"),
            cvcuda.Tensor([5, 21, 10, 3], np.uint8, "NHWC"),
            cvcuda.Tensor([5, 21, 10, 1], np.uint8, "NHWC"),
            4,
        ),
    ],
)
def test_op_composite(foreground, background, fgMask, outChannels):
    out = cvcuda.composite(foreground, background, fgMask, outChannels)
    assert out.layout == foreground.layout
    assert out.shape[-1] == outChannels
    if out.layout == "NHWC":
        assert out.shape[0:3] == foreground.shape[0:3]
    if out.layout == "HWC":
        assert out.shape[0:2] == foreground.shape[0:2]
    assert out.dtype == foreground.dtype

    stream = cvcuda.Stream()

    out_shape = foreground.shape
    out_shape[-1] = outChannels
    out = cvcuda.Tensor(out_shape, foreground.dtype, foreground.layout)
    tmp = cvcuda.composite_into(
        foreground=foreground,
        dst=out,
        background=background,
        fgmask=fgMask,
        stream=stream,
    )
    assert tmp is out
    assert out.layout == foreground.layout
    assert out.shape[-1] == outChannels
    if out.layout == "NHWC":
        assert out.shape[0:3] == foreground.shape[0:3]
    if out.layout == "HWC":
        assert out.shape[0:2] == foreground.shape[0:2]
    assert out.dtype == foreground.dtype


@t.mark.parametrize(
    "nimages, max_size, outChannels",
    [
        (
            5,
            (10, 20),
            3,
        ),
        (
            8,
            (10, 20),
            4,
        ),
    ],
)
def test_op_compositevarshape(nimages, max_size, outChannels):
    foreground = util.create_image_batch(
        nimages, cvcuda.Format.RGB8, max_size=max_size, max_random=255, rng=RNG
    )

    background = util.clone_image_batch(foreground)
    fgMask = util.clone_image_batch(foreground, img_format=cvcuda.Format.U8)

    out = cvcuda.composite(foreground, background, fgMask)
    assert len(out) == len(foreground)
    assert out.capacity == foreground.capacity

    stream = cvcuda.cuda.Stream()

    if outChannels == 3:
        out = util.clone_image_batch(foreground)
    if outChannels == 4:
        out = util.clone_image_batch(foreground, img_format=cvcuda.Format.RGBA8)
    tmp = cvcuda.composite_into(
        foreground=foreground,
        dst=out,
        background=background,
        fgmask=fgMask,
        stream=stream,
    )

    assert tmp is out
    assert len(out) == len(foreground)
    assert out.capacity == foreground.capacity



# SPDX-FileCopyrightText: Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cvcuda
import pytest as t
import numpy as np
import util

RNG = np.random.default_rng(0)


@t.mark.parametrize(
    "input, diameter, sigma_color, sigma_space, border",
    [
        (
            cvcuda.Tensor([5, 9, 9, 4], np.uint8, "NHWC"),
            9,
            1,
            1,
            cvcuda.Border.CONSTANT,
        ),
        (
            cvcuda.Tensor([9, 9, 3], np.uint8, "HWC"),
            7,
            3,
            10,
            cvcuda.Border.WRAP,
        ),
        (
            cvcuda.Tensor([5, 21, 21, 4], np.uint8, "NHWC"),
            6,
            15,
            9,
            cvcuda.Border.REPLICATE,
        ),
        (
            cvcuda.Tensor([21, 21, 3], np.uint8, "HWC"),
            12,
            2,
            5,
            cvcuda.Border.REFLECT,
        ),
    ],
)
def test_op_bilateral_filter(input, diameter, sigma_color, sigma_space, border):
    out = cvcuda.bilateral_filter(
        input, diameter, sigma_color, sigma_space, border=border
    )
    assert out.layout == input.layout
    assert out.shape == input.shape
    assert out.dtype == input.dtype

    stream = cvcuda.Stream()
    out = cvcuda.Tensor(input.shape, input.dtype, input.layout)
    tmp = cvcuda.bilateral_filter_into(
        src=input,
        dst=out,
        diameter=diameter,
        sigma_color=sigma_color,
        sigma_space=sigma_space,
        border=border,
        stream=stream,
    )
    assert tmp is out
    assert out.layout == input.layout
    assert out.shape == input.shape
    assert out.dtype == input.dtype


@t.mark.parametrize(
    "nimages, format, max_size, max_pixel, max_diameter, max_sc, max_ss, border",
    [
        (
            5,
            cvcuda.Format.RGB8,
            (16, 23),
            128.0,
            12,
            2,
            5,
            cvcuda.Border.REFLECT,
        ),
        (
            5,
            cvcuda.Format.RGB8,
            (16, 23),
            256.0,
            6,
            15,
            9,
            cvcuda.Border.REPLICATE,
        ),
        (
            5,
            cvcuda.Format.RGB8,
            (16, 23),
            256.0,
            7,
            3,
            10,
            cvcuda.Border.WRAP,
        ),
        (
            4,
            cvcuda.Format.RGB8,
            (11, 23),
            256.0,
            9,
            1,
            1,
            cvcuda.Border.CONSTANT,
        ),
    ],
)
def test_op_bilateral_filtervarshape(
    nimages,
    format,
    max_size,
    max_pixel,
    max_diameter,
    max_sc,
    max_ss,
    border,
):

    input = util.create_image_batch(
        nimages, format, max_size=max_size, max_random=max_pixel, rng=RNG
    )

    diameter = util.create_tensor(
        (nimages), np.int32, "N", max_random=max_diameter, rng=RNG
    )

    sigma_color = util.create_tensor(
        (nimages), np.float32, "N", max_random=max_sc, rng=RNG
    )

    sigma_space = util.create_tensor(
        (nimages), np.float32, "N", max_random=max_ss, rng=RNG
    )
    out = cvcuda.bilateral_filter(
        input, diameter, sigma_color, sigma_space, border=border
    )

    assert len(out) == len(input)
    assert out.capacity == input.capacity
    assert out.uniqueformat == input.uniqueformat
    assert out.maxsize == input.maxsize

    stream = cvcuda.Stream()

    out = util.clone_image_batch(input)

    tmp = cvcuda.bilateral_filter_into(
        src=input,
        dst=out,
        diameter=diameter,
        sigma_color=sigma_color,
        sigma_space=sigma_space,
        border=border,
        stream=stream,
    )
    assert tmp is out
    assert len(out) == len(input)
    assert out.capacity == input.capacity
    assert out.uniqueformat == input.uniqueformat
    assert out.maxsize == input.maxsize



# SPDX-FileCopyrightText: Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cvcuda
import pytest as t
import numpy as np


@t.mark.parametrize(
    "input, top, bottom, left, right, border_mode, border_value",
    [
        (
            cvcuda.Tensor([5, 16, 23, 4], np.uint8, "NHWC"),
            1,
            2,
            3,
            4,
            cvcuda.Border.CONSTANT,
            [0],
        ),
        (
            cvcuda.Tensor([5, 16, 23, 4], np.uint8, "NHWC"),
            1,
            2,
            3,
            4,
            cvcuda.Border.CONSTANT,
            [12, 3, 4, 55],
        ),
        (
            cvcuda.Tensor([16, 23, 4], np.uint8, "HWC"),
            2,
            2,
            2,
            2,
            cvcuda.Border.WRAP,
            [0],
        ),
        (
            cvcuda.Tensor([16, 23, 4], np.uint8, "HWC"),
            2,
            2,
            2,
            2,
            cvcuda.Border.WRAP,
            [0],
        ),
        (
            cvcuda.Tensor([16, 23, 3], np.uint8, "HWC"),
            10,
            12,
            35,
            18,
            cvcuda.Border.REPLICATE,
            [0],
        ),
        (
            cvcuda.Tensor([16, 23, 1], np.float32, "HWC"),
            11,
            1,
            20,
            3,
            cvcuda.Border.REFLECT,
            [0],
        ),
        (
            cvcuda.Tensor([16, 23, 3], np.float32, "HWC"),
            11,
            1,
            20,
            3,
            cvcuda.Border.REFLECT101,
            [0],
        ),
    ],
)
def test_op_copymakeborder(input, top, bottom, left, right, border_mode, border_value):
    out_shape = [i for i in input.shape]
    cdim = len(out_shape) - 1
    out_shape[cdim - 2] += top + bottom
    out_shape[cdim - 1] += left + right
    out = cvcuda.copymakeborder(input, top=top, bottom=bottom, left=left, right=right)
    assert out.layout == input.layout
    assert out.shape == out_shape
    assert out.dtype == input.dtype

    stream = cvcuda.Stream()
    out = cvcuda.Tensor(out_shape, input.dtype, input.layout)
    tmp = cvcuda.copymakeborder_into(
        src=input,
        dst=out,
        top=top,
        left=left,
        border_mode=border_mode,
        border_value=border_value,
        stream=stream,
    )
    assert tmp is out
    assert out.layout == input.layout
    assert out.shape == out_shape
    assert out.dtype == input.dtype



# SPDX-FileCopyrightText: Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import nvcv
import numpy as np
import torch
import util


RNG = np.random.default_rng(0)


def test_create_tensor_uint16():
    tensor = util.create_tensor([10], np.uint16, None, max_random=(65535))
    assert tensor.dtype == np.uint16


def test_create_tensor_odd():
    tensor = util.create_tensor(
        [10], np.uint8, None, max_random=255, rng=RNG, transform_dist=util.dist_odd
    )
    h_data = torch.as_tensor(tensor.cuda(), device="cuda").cpu()
    assert all([bool(val % 2 == 1) for val in h_data])


def test_create_image():
    image = util.create_image((11, 13), nvcv.Format.RGB8, max_random=100, rng=RNG)
    assert image.size == (11, 13)
    assert image.format == nvcv.Format.RGB8
    assert all([bool(val < 100) for val in np.array(image.cpu()).flatten()])


def test_create_image_batch():
    batch = util.create_image_batch(
        19, nvcv.Format.RGBA8, max_size=(3, 3), max_random=33, rng=RNG
    )
    assert batch.capacity == 19
    assert len(batch) == 19
    assert batch.uniqueformat == nvcv.Format.RGBA8
    assert batch.maxsize == (3, 3)
    assert all(
        [bool(val < 33) for img in batch for val in np.array(img.cpu()).flatten()]
    )


def test_clone_image_batch():
    batch1 = util.create_image_batch(17, nvcv.Format.RGB8, size=(4, 4))
    batch2 = util.clone_image_batch(batch1)
    assert batch2.capacity == 17
    assert len(batch2) == 17
    assert batch2.uniqueformat == nvcv.Format.RGB8
    assert batch2.maxsize == (4, 4)



# SPDX-FileCopyrightText: Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest as t
import nvcv
import numpy as np
import math
from numba import cuda


def test_image_creation_works():
    img = nvcv.Image((7, 5), nvcv.Format.NV12)
    assert img.width == 7
    assert img.height == 5
    assert img.size == (7, 5)
    assert img.format == nvcv.Format.NV12


def test_image_creation_arg_keywords():
    img = nvcv.Image(size=(7, 5), format=nvcv.Format.NV12)
    assert img.width == 7
    assert img.height == 5
    assert img.size == (7, 5)
    assert img.format == nvcv.Format.NV12


buffmt_common = [
    # packed formats
    ([5, 7, 1], np.uint8, nvcv.Format.U8),
    ([5, 7, 1], np.uint8, nvcv.Format.U8),
    ([5, 7, 1], np.uint8, nvcv.Format.U8),
    ([5, 7], np.uint8, nvcv.Format.U8),
    ([5, 7, 1], np.int8, nvcv.Format.S8),
    ([5, 7, 1], np.uint16, nvcv.Format.U16),
    ([5, 7, 1], np.int16, nvcv.Format.S16),
    ([5, 7, 2], np.int16, nvcv.Format._2S16),
    ([5, 7, 1], np.float32, nvcv.Format.F32),
    ([5, 7, 1], np.float64, nvcv.Format.F64),
    ([5, 7, 2], np.float32, nvcv.Format._2F32),
    ([5, 7, 3], np.uint8, nvcv.Format.RGB8),
    ([5, 7, 4], np.uint8, nvcv.Format.RGBA8),
    ([1, 5, 7], np.uint8, nvcv.Format.U8),
    ([1, 5, 7, 4], np.uint8, nvcv.Format.RGBA8),
    ([5, 7], np.csingle, nvcv.Format._2F32),
]


@t.mark.parametrize("shape,dt,format", buffmt_common)
def test_wrap_host_buffer_infer_imgformat(shape, dt, format):
    img = nvcv.Image(np.ndarray(shape, dt))
    assert img.width == 7
    assert img.height == 5
    assert img.format == format

    img = nvcv.as_image(cuda.device_array(shape, dt))
    assert img.width == 7
    assert img.height == 5
    assert img.format == format


@t.mark.parametrize(
    "shape,dt,format",
    buffmt_common
    + [
        ([5, 7, 1], np.uint8, nvcv.Format.Y8),
        ([5, 7, 3], np.uint8, nvcv.Format.BGR8),
        ([5, 7, 4], np.uint8, nvcv.Format.BGRA8),
    ],
)
def test_wrap_host_buffer_explicit_format(shape, dt, format):
    img = nvcv.Image(np.ndarray(shape, dt), format)
    assert img.width == 7
    assert img.height == 5
    assert img.format == format

    img = nvcv.as_image(cuda.device_array(shape, dt), format)
    assert img.width == 7
    assert img.height == 5
    assert img.format == format


buffmt2_common = [
    # packed formats
    (
        [([6, 8], np.uint8), ([3, 4, 2], np.uint8)],
        nvcv.Format.NV12_ER,
    )
]


@t.mark.parametrize("buffers,format", buffmt2_common)
@t.mark.xfail  # wrapping of multiple cuda buffers is failing
def test_wrap_host_buffer_infer_imgformat_multiple_planes(buffers, format):
    img = nvcv.Image([np.ndarray(buf[0], buf[1]) for buf in buffers])
    assert img.width == 8
    assert img.height == 6
    assert img.format == format

    img = nvcv.as_image([cuda.device_array(buf[0], buf[1]) for buf in buffers])
    assert img.width == 8
    assert img.height == 6
    assert img.format == format


@t.mark.parametrize("buffers,format", buffmt2_common)
@t.mark.xfail  # wrapping of multiple cuda buffers is failing
def test_wrap_host_buffer_explicit_format2(buffers, format):
    img = nvcv.Image([np.ndarray(buf[0], buf[1]) for buf in buffers], format)
    assert img.width == 8
    assert img.height == 6
    assert img.format == format

    img = nvcv.as_image([cuda.device_array(buf[0], buf[1]) for buf in buffers], format)
    assert img.width == 8
    assert img.height == 6
    assert img.format == format


@t.mark.parametrize(
    "shape,dt,planes,height,width,channels",
    [
        ([2, 7, 6], np.uint8, 2, 7, 6, 2),
        ([1, 2, 7, 6], np.uint8, 2, 7, 6, 2),
        ([2, 7, 3], np.uint8, 1, 2, 7, 3),
        ([1, 7, 3], np.uint8, 1, 1, 7, 3),
        ([7, 3], np.uint8, 1, 7, 3, 1),
        ([7, 1], np.uint8, 1, 7, 1, 1),
        ([1, 3], np.uint8, 1, 1, 3, 1),
        ([1, 1], np.uint8, 1, 1, 1, 1),
        ([5, 7, 3], np.uint8, 1, 5, 7, 3),
    ],
)
def test_wrap_host_buffer_infer_format_geometry(
    shape, dt, planes, height, width, channels
):
    img = nvcv.Image(np.ndarray(shape, dt))
    assert img.width == width
    assert img.height == height
    assert img.format.planes == planes
    assert img.format.channels == channels

    img = nvcv.as_image(cuda.device_array(shape, dt))
    assert img.width == width
    assert img.height == height
    assert img.format.planes == planes
    assert img.format.channels == channels


def test_wrap_host_buffer_arg_keywords():
    img = nvcv.Image(buffer=np.ndarray([5, 7], np.float32), format=nvcv.Format.F32)
    assert img.size == (7, 5)
    assert img.format == nvcv.Format.F32

    img = nvcv.as_image(
        buffer=cuda.device_array([5, 7], np.float32), format=nvcv.Format.F32
    )
    assert img.size == (7, 5)
    assert img.format == nvcv.Format.F32


def test_wrap_host_buffer_infer_format_arg_keywords():
    img = nvcv.Image(buffer=np.ndarray([5, 7], np.float32))
    assert img.size == (7, 5)
    assert img.format == nvcv.Format.F32

    img = nvcv.as_image(buffer=cuda.device_array([5, 7], np.float32))
    assert img.size == (7, 5)
    assert img.format == nvcv.Format.F32


def test_wrap_host_image_with_format__buffer_has_unsupported_type():
    with t.raises(ValueError):
        nvcv.Image(np.array([1 + 2j, 4 + 7j]), nvcv.Format._2F32)


def test_wrap_host_image__buffer_has_unsupported_type():
    with t.raises(ValueError):
        nvcv.Image(np.array([1 + 2j, 4 + 7j]))


def test_wrap_host_image__format_and_buffer_type_mismatch():
    with t.raises(ValueError):
        nvcv.Image(np.array([1.4, 2.85]), nvcv.Format.U8)


def test_wrap_host_image__only_pitch_linear():
    with t.raises(ValueError):
        nvcv.Image(np.ndarray([6, 4], np.uint8), nvcv.Format.Y8_BL)


def test_wrap_host_image__css_with_one_plane_failure():
    with t.raises(ValueError):
        nvcv.Image(np.ndarray([6, 4], np.uint8), nvcv.Format.NV12)


@t.mark.parametrize(
    "shape",
    [
        (5, 3, 4),  # Buffer shape HCW not supported
        (5, 7, 4),  # Buffer shape doesn't correspond to image format
    ],
)
def test_wrap_host_image_with_format__invalid_shape(shape):
    with t.raises(ValueError):
        nvcv.Image(np.ndarray(shape, np.uint8), nvcv.Format.RGB8)


@t.mark.parametrize(
    "shape",
    [
        # When buffer's number of dimensions is 4, first dimension must be 1, not 2
        (
            2,
            3,
            4,
            5,
        ),
        # Number of dimensions must be between 1 and 4, not 5
        (1, 1, 15, 7, 1),
        # Number of dimensions must be between 1 and 4, not 0
        (0,),
        # Buffer shape not supported
        (8, 7, 9),
    ],
)
def test_wrap_host_invalid_dims(shape):
    with t.raises(ValueError):
        nvcv.Image(np.ndarray(shape))


@t.mark.parametrize(
    "s",
    [
        # Fastest changing dimension must be packed, i.e.,
        # have stride equal to 1 bytes(s), not 2
        (2 * 2, 2),
        # Buffer strides must all be >= 0
        (0, 1),
    ],
)
def test_wrap_host_invalid_strides(s):
    with t.raises(ValueError):
        nvcv.Image(
            np.ndarray(
                shape=(3, 2), strides=s, buffer=bytearray(s[0] * 3), dtype=np.uint8
            )
        )


@t.mark.parametrize(
    "shapes",
    [
        # When wrapping multiple buffers, buffers with 4
        # dimensions must have first dimension == 1, not 2
        [
            (3, 4),
            (2, 2, 3, 1),
        ],
        # Number of buffer#1's dimensions must be
        # between 1 and 4, not 5
        [
            (3, 4),
            (5, 2, 2, 3, 1),
        ],
    ],
)
def test_wrap_host_multiplane_invalid_dims(shapes):
    buffers = []
    for shape in shapes:
        buffers.append(np.ndarray(shape, np.uint8))

    with t.raises(ValueError):
        nvcv.Image(buffers)


def test_image_wrap_invalid_cuda_buffer():
    class NonCudaMemory(object):
        pass

    obj = NonCudaMemory()
    obj.__cuda_array_interface__ = {
        "shape": (1, 1),
        "typestr": "i",
        "data": (419, True),
        "version": 3,
    }

    with t.raises(RuntimeError):
        nvcv.as_image(obj)


@cuda.jit
def set_value(buffer):
    z, y, x = cuda.grid(3)
    if z < buffer.shape[0] and y < buffer.shape[1] and x < buffer.shape[2]:
        buffer[z, y, x] = (z * buffer.shape[1] + y) * buffer.shape[2] + x


@t.mark.parametrize(
    "size,format,layout,out_dtype, out_shape, simple_layout",
    [
        ((257, 231), nvcv.Format.U8, None, np.uint8, [231, 257], "HWC"),
        ((257, 231), nvcv.Format.U8, "HWC", np.uint8, [231, 257, 1], "HWC"),
        ((257, 231), nvcv.Format.U8, "CHW", np.uint8, [1, 231, 257], "CHW"),
        (
            (257, 231),
            nvcv.Format.U8,
            "xyCrodHlimaWab",
            np.uint8,
            [1, 1, 1, 1, 1, 1, 231, 1, 1, 1, 1, 257, 1, 1],
            "CHW",
        ),
        ((257, 231), nvcv.Format.RGBAf32, None, np.float32, [231, 257, 4], "HWC"),
        ((257, 231), nvcv.Format.RGBA8, "HWC", np.uint8, [231, 257, 4], "HWC"),
        ((257, 231), nvcv.Format.RGBA8p, None, np.uint8, [4, 231, 257], "CHW"),
        ((257, 231), nvcv.Format.RGBAf32p, "CHW", np.float32, [4, 231, 257], "CHW"),
        (
            (258, 232),
            nvcv.Format.NV12,
            None,
            [np.uint8, np.uint8],
            [[232, 258, 1], [232 // 2, 258 // 2, 2]],
            "HWC",
        ),
        (
            (258, 232),
            nvcv.Format.NV12,
            "HWC",
            [np.uint8, np.uint8],
            [[232, 258, 1], [232 // 2, 258 // 2, 2]],
            "HWC",
        ),
        # For YUYV and friends things get a bit funky
        ((258, 232), nvcv.Format.YUYV, None, np.uint8, [232, 258, 2], "HWC"),
        ((258, 232), nvcv.Format.YUYV, "HWC", np.uint8, [232, 258, 2], "HWC"),
    ],
)
def test_image_export_cuda_buffer(
    size, format, layout, out_dtype, out_shape, simple_layout
):
    img = nvcv.Image(size, format)

    mem = img.cuda(layout)
    if type(mem) is list:
        for i in range(0, len(mem)):
            assert mem[i].dtype == out_dtype[i]
            assert mem[i].shape == out_shape[i]
    else:
        assert mem.dtype == out_dtype
        assert mem.shape == out_shape

    # Make sure buffer cache works
    assert img.cuda(layout) is mem
    if layout is not None:
        newmem = img.cuda()
        assert newmem is not mem
        assert newmem is img.cuda()
        assert newmem is not img.cuda(layout)

    cuda_buffer = img.cuda(simple_layout)
    if type(cuda_buffer) is not list:
        cuda_buffer = [cuda_buffer]

    # Write values in it on CUDA side
    block = (32, 4, 4)
    for buf in cuda_buffer:
        grid = (
            math.ceil(buf.shape[0] / block[0]),
            math.ceil(buf.shape[1] / block[1]),
            math.ceil(buf.shape[2] / block[2]),
        )
        set_value[grid, block](buf)

    cuda.synchronize()

    # Get values back on cpu
    host_buffer = img.cpu(simple_layout)
    if type(host_buffer) is not list:
        host_buffer = [host_buffer]

    if type(out_dtype) is not list:
        out_dtype = [out_dtype]

    # compare to see if they are correct
    for b in range(0, len(host_buffer)):
        buf = host_buffer[b]
        for z in range(0, buf.shape[0]):
            for y in range(0, buf.shape[1]):
                for x in range(0, buf.shape[2]):
                    assert buf[z, y, x] == out_dtype[b](
                        (z * buf.shape[1] + y) * buf.shape[2] + x
                    )


def test_image_zeros():
    img = nvcv.Image.zeros((67, 34), nvcv.Format.F32)
    assert (img.cpu() == np.zeros([34, 67], np.float32)).all()



import sys

import pytest

np = pytest.importorskip("numpy")
eigen_tensor = pytest.importorskip("pybind11_tests.eigen_tensor")
submodules = [eigen_tensor.c_style, eigen_tensor.f_style]
try:
    import eigen_tensor_avoid_stl_array as avoid

    submodules += [avoid.c_style, avoid.f_style]
except ImportError as e:
    # Ensure config, build, toolchain, etc. issues are not masked here:
    raise RuntimeError(
        "import eigen_tensor_avoid_stl_array FAILED, while "
        "import pybind11_tests.eigen_tensor succeeded. "
        "Please ensure that "
        "test_eigen_tensor.cpp & "
        "eigen_tensor_avoid_stl_array.cpp "
        "are built together (or both are not built if Eigen is not available)."
    ) from e

tensor_ref = np.empty((3, 5, 2), dtype=np.int64)

for i in range(tensor_ref.shape[0]):
    for j in range(tensor_ref.shape[1]):
        for k in range(tensor_ref.shape[2]):
            tensor_ref[i, j, k] = i * (5 * 2) + j * 2 + k

indices = (2, 3, 1)


@pytest.fixture(autouse=True)
def cleanup():
    for module in submodules:
        module.setup()

    yield

    for module in submodules:
        assert module.is_ok()


def test_import_avoid_stl_array():
    pytest.importorskip("eigen_tensor_avoid_stl_array")
    assert len(submodules) == 4


def assert_equal_tensor_ref(mat, writeable=True, modified=None):
    assert mat.flags.writeable == writeable

    copy = np.array(tensor_ref)
    if modified is not None:
        copy[indices] = modified

    np.testing.assert_array_equal(mat, copy)


@pytest.mark.parametrize("m", submodules)
@pytest.mark.parametrize("member_name", ["member", "member_view"])
def test_reference_internal(m, member_name):

    if not hasattr(sys, "getrefcount"):
        pytest.skip("No reference counting")
    foo = m.CustomExample()
    counts = sys.getrefcount(foo)
    mem = getattr(foo, member_name)
    assert_equal_tensor_ref(mem, writeable=False)
    new_counts = sys.getrefcount(foo)
    assert new_counts == counts + 1
    assert_equal_tensor_ref(mem, writeable=False)
    del mem
    assert sys.getrefcount(foo) == counts


assert_equal_funcs = [
    "copy_tensor",
    "copy_fixed_tensor",
    "copy_const_tensor",
    "move_tensor_copy",
    "move_fixed_tensor_copy",
    "take_tensor",
    "take_fixed_tensor",
    "reference_tensor",
    "reference_tensor_v2",
    "reference_fixed_tensor",
    "reference_view_of_tensor",
    "reference_view_of_tensor_v3",
    "reference_view_of_tensor_v5",
    "reference_view_of_fixed_tensor",
]

assert_equal_const_funcs = [
    "reference_view_of_tensor_v2",
    "reference_view_of_tensor_v4",
    "reference_view_of_tensor_v6",
    "reference_const_tensor",
    "reference_const_tensor_v2",
]


@pytest.mark.parametrize("m", submodules)
@pytest.mark.parametrize("func_name", assert_equal_funcs + assert_equal_const_funcs)
def test_convert_tensor_to_py(m, func_name):
    writeable = func_name in assert_equal_funcs
    assert_equal_tensor_ref(getattr(m, func_name)(), writeable=writeable)


@pytest.mark.parametrize("m", submodules)
def test_bad_cpp_to_python_casts(m):

    with pytest.raises(
        RuntimeError, match="Cannot use reference internal when there is no parent"
    ):
        m.reference_tensor_internal()

    with pytest.raises(RuntimeError, match="Cannot move from a constant reference"):
        m.move_const_tensor()

    with pytest.raises(
        RuntimeError, match="Cannot take ownership of a const reference"
    ):
        m.take_const_tensor()

    with pytest.raises(
        RuntimeError,
        match="Invalid return_value_policy for Eigen Map type, must be either reference or reference_internal",
    ):
        m.take_view_tensor()


@pytest.mark.parametrize("m", submodules)
def test_bad_python_to_cpp_casts(m):

    with pytest.raises(
        TypeError, match=r"^round_trip_tensor\(\): incompatible function arguments"
    ):
        m.round_trip_tensor(np.zeros((2, 3)))

    with pytest.raises(TypeError, match=r"^Cannot cast array data from dtype"):
        m.round_trip_tensor(np.zeros(dtype=np.str_, shape=(2, 3, 1)))

    with pytest.raises(
        TypeError,
        match=r"^round_trip_tensor_noconvert\(\): incompatible function arguments",
    ):
        m.round_trip_tensor_noconvert(tensor_ref)

    assert_equal_tensor_ref(
        m.round_trip_tensor_noconvert(tensor_ref.astype(np.float64))
    )

    if m.needed_options == "F":
        bad_options = "C"
    else:
        bad_options = "F"
    # Shape, dtype and the order need to be correct for a TensorMap cast
    with pytest.raises(
        TypeError, match=r"^round_trip_view_tensor\(\): incompatible function arguments"
    ):
        m.round_trip_view_tensor(
            np.zeros((3, 5, 2), dtype=np.float64, order=bad_options)
        )

    with pytest.raises(
        TypeError, match=r"^round_trip_view_tensor\(\): incompatible function arguments"
    ):
        m.round_trip_view_tensor(
            np.zeros((3, 5, 2), dtype=np.float32, order=m.needed_options)
        )

    with pytest.raises(
        TypeError, match=r"^round_trip_view_tensor\(\): incompatible function arguments"
    ):
        m.round_trip_view_tensor(
            np.zeros((3, 5), dtype=np.float64, order=m.needed_options)
        )

    with pytest.raises(
        TypeError, match=r"^round_trip_view_tensor\(\): incompatible function arguments"
    ):
        temp = np.zeros((3, 5, 2), dtype=np.float64, order=m.needed_options)
        m.round_trip_view_tensor(
            temp[:, ::-1, :],
        )

    with pytest.raises(
        TypeError, match=r"^round_trip_view_tensor\(\): incompatible function arguments"
    ):
        temp = np.zeros((3, 5, 2), dtype=np.float64, order=m.needed_options)
        temp.setflags(write=False)
        m.round_trip_view_tensor(temp)


@pytest.mark.parametrize("m", submodules)
def test_references_actually_refer(m):

    a = m.reference_tensor()
    temp = a[indices]
    a[indices] = 100
    assert_equal_tensor_ref(m.copy_const_tensor(), modified=100)
    a[indices] = temp
    assert_equal_tensor_ref(m.copy_const_tensor())

    a = m.reference_view_of_tensor()
    a[indices] = 100
    assert_equal_tensor_ref(m.copy_const_tensor(), modified=100)
    a[indices] = temp
    assert_equal_tensor_ref(m.copy_const_tensor())


@pytest.mark.parametrize("m", submodules)
def test_round_trip(m):

    assert_equal_tensor_ref(m.round_trip_tensor(tensor_ref))

    with pytest.raises(TypeError, match="^Cannot cast array data from"):
        assert_equal_tensor_ref(m.round_trip_tensor2(tensor_ref))

    assert_equal_tensor_ref(m.round_trip_tensor2(np.array(tensor_ref, dtype=np.int32)))
    assert_equal_tensor_ref(m.round_trip_fixed_tensor(tensor_ref))
    assert_equal_tensor_ref(m.round_trip_aligned_view_tensor(m.reference_tensor()))

    copy = np.array(tensor_ref, dtype=np.float64, order=m.needed_options)
    assert_equal_tensor_ref(m.round_trip_view_tensor(copy))
    assert_equal_tensor_ref(m.round_trip_view_tensor_ref(copy))
    assert_equal_tensor_ref(m.round_trip_view_tensor_ptr(copy))
    copy.setflags(write=False)
    assert_equal_tensor_ref(m.round_trip_const_view_tensor(copy))

    np.testing.assert_array_equal(
        tensor_ref[:, ::-1, :], m.round_trip_tensor(tensor_ref[:, ::-1, :])
    )

    assert m.round_trip_rank_0(np.float64(3.5)) == 3.5
    assert m.round_trip_rank_0(3.5) == 3.5

    with pytest.raises(
        TypeError,
        match=r"^round_trip_rank_0_noconvert\(\): incompatible function arguments",
    ):
        m.round_trip_rank_0_noconvert(np.float64(3.5))

    with pytest.raises(
        TypeError,
        match=r"^round_trip_rank_0_noconvert\(\): incompatible function arguments",
    ):
        m.round_trip_rank_0_noconvert(3.5)

    with pytest.raises(
        TypeError, match=r"^round_trip_rank_0_view\(\): incompatible function arguments"
    ):
        m.round_trip_rank_0_view(np.float64(3.5))

    with pytest.raises(
        TypeError, match=r"^round_trip_rank_0_view\(\): incompatible function arguments"
    ):
        m.round_trip_rank_0_view(3.5)


@pytest.mark.parametrize("m", submodules)
def test_round_trip_references_actually_refer(m):

    # Need to create a copy that matches the type on the C side
    copy = np.array(tensor_ref, dtype=np.float64, order=m.needed_options)
    a = m.round_trip_view_tensor(copy)
    temp = a[indices]
    a[indices] = 100
    assert_equal_tensor_ref(copy, modified=100)
    a[indices] = temp
    assert_equal_tensor_ref(copy)


@pytest.mark.parametrize("m", submodules)
def test_doc_string(m, doc):
    assert (
        doc(m.copy_tensor) == "copy_tensor() -> numpy.ndarray[numpy.float64[?, ?, ?]]"
    )
    assert (
        doc(m.copy_fixed_tensor)
        == "copy_fixed_tensor() -> numpy.ndarray[numpy.float64[3, 5, 2]]"
    )
    assert (
        doc(m.reference_const_tensor)
        == "reference_const_tensor() -> numpy.ndarray[numpy.float64[?, ?, ?]]"
    )

    order_flag = f"flags.{m.needed_options.lower()}_contiguous"
    assert doc(m.round_trip_view_tensor) == (
        f"round_trip_view_tensor(arg0: numpy.ndarray[numpy.float64[?, ?, ?], flags.writeable, {order_flag}])"
        + f" -> numpy.ndarray[numpy.float64[?, ?, ?], flags.writeable, {order_flag}]"
    )
    assert doc(m.round_trip_const_view_tensor) == (
        f"round_trip_const_view_tensor(arg0: numpy.ndarray[numpy.float64[?, ?, ?], {order_flag}])"
        + " -> numpy.ndarray[numpy.float64[?, ?, ?]]"
    )



# SPDX-FileCopyrightText: Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import nvcv
import pytest as t
import numpy as np


@t.mark.parametrize(
    "type,dt",
    [
        (nvcv.Type.U8, np.uint8),
        (nvcv.Type.U8, np.dtype(np.uint8)),
        (nvcv.Type.S8, np.int8),
        (nvcv.Type.U16, np.uint16),
        (nvcv.Type.S16, np.int16),
        (nvcv.Type.U32, np.uint32),
        (nvcv.Type.S32, np.int32),
        (nvcv.Type.U64, np.uint64),
        (nvcv.Type.S64, np.int64),
        (nvcv.Type.F32, np.float32),
        (nvcv.Type.F64, np.float64),
        (nvcv.Type._2F32, np.complex64),
        (nvcv.Type._2F64, np.complex128),
        (nvcv.Type._3S8, np.dtype("3i1")),
        (nvcv.Type._4S32, np.dtype("4i")),
    ],
)
def test_pixtype_dtype(type, dt):
    assert type == dt

    t = nvcv.Type(dt)
    assert type == t
    assert dt == t


@t.mark.parametrize("dt", [np.dtype([("f1", np.uint64), ("f2", np.int32)]), "invalid"])
def test_pixtype_dtype_conv_error(dt):
    with t.raises(TypeError):
        nvcv.Type(dt)



# SPDX-FileCopyrightText: Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Semantic Segmentation Python sample

The semantic segmentation sample uses DeepLabv3 model from the torchvision
repository and shows the usage of CVCUDA by implementing a complete end-to-end
pipeline which can read images from the disk, pre-process them, run the inference
on them and save the overlay outputs back to the disk. This sample also gives an
overview of the interoperability of PyTorch and TensorRT with CVCUDA tensors and
operators.
"""

# docs_tag: begin_python_imports
import os
import sys
import glob
import argparse
import torch
import torchnvjpeg
import torchvision.transforms.functional as F
from torchvision.models import segmentation as segmentation_models
import numpy as np
import cvcuda
import tensorrt as trt

# Bring the commons folder from the samples directory into our path so that
# we can import modules from it.
common_dir = os.path.join(
    os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))),
    "common",
    "python",
)
sys.path.insert(0, common_dir)
from trt_utils import convert_onnx_to_tensorrt, setup_tensort_bindings  # noqa: E402

# docs_tag: end_python_imports


# docs_tag: class_def
class SemanticSegmentationSample:
    def __init__(
        self,
        input_path,
        results_dir,
        visualization_class_name,
        batch_size,
        target_img_height,
        target_img_width,
        device_id,
        backend,
    ):
        # docs_tag: begin_class_init
        self.input_path = input_path
        self.results_dir = results_dir
        self.visualization_class_name = visualization_class_name
        self.batch_size = batch_size
        self.target_img_height = target_img_height
        self.target_img_width = target_img_width
        self.device_id = device_id
        self.backend = backend
        self.class_to_idx_dict = None

        if self.backend not in ["pytorch", "tensorrt"]:
            print(
                "Invalid backend option specified: %s. "
                "Currently supports: pytorch, tensorrt" % self.backend
            )
            exit(1)
        # docs_tag: end_class_init

        # docs_tag: begin_data_read
        # Start by parsing the input_path expression first.
        if os.path.isfile(self.input_path):
            # Read the input image file.
            self.file_names = [self.input_path] * self.batch_size
            # Then create a dummy list with the data from the same file to simulate a
            # batch.
            self.data = [open(path, "rb").read() for path in self.file_names]

        elif os.path.isdir(self.input_path):
            # It is a directory. Grab all the images from it.
            self.file_names = glob.glob(os.path.join(self.input_path, "*.jpg"))
            self.data = [open(path, "rb").read() for path in self.file_names]
            print("Read a total of %d JPEG images." % len(self.data))

        else:
            print(
                "Input path not found. "
                "It is neither a valid JPEG file nor a directory: %s" % self.input_path
            )
            exit(1)
        # docs_tag: end_data_read

        # docs_tag: begin_input_validation
        if not os.path.isdir(self.results_dir):
            print("Output directory not found: %s" % self.results_dir)
            exit(1)

        if self.batch_size <= 0:
            print("batch_size must be a value >=1.")
            exit(1)

        if self.target_img_height < 10:
            print("target_img_height must be a value >=10.")
            exit(1)

        if self.target_img_width < 10:
            print("target_img_width must be a value >=10.")
            exit(1)
        # docs_tag: end_input_validation

    # docs_tag: setup_model_def
    def setup_model(self):
        # Setup the model and a few more things depending on the type of backend.
        # docs_tag: begin_setup_pytorch
        if self.backend == "pytorch":
            # Fetch the segmentation index to class name information from the weights
            # meta properties.
            torch_model = segmentation_models.deeplabv3_resnet101
            weights_info = segmentation_models.DeepLabV3_ResNet101_Weights

            weights = weights_info.DEFAULT
            self.class_to_idx_dict = {
                cls: idx for (idx, cls) in enumerate(weights.meta["categories"])
            }

            if self.visualization_class_name not in self.class_to_idx_dict:
                print(
                    "Requested segmentation class '%s' is not supported by the "
                    "DeepLabV3 model." % self.visualization_class_name
                )
                print(
                    "All supported class names are: %s"
                    % ", ".join(self.class_to_idx_dict.keys())
                )
                exit(1)

            # Inference uses PyTorch to run a segmentation model on the pre-processed
            # input and outputs the segmentation masks.
            model = torch_model(weights=weights)
            model.cuda(self.device_id)
            model.eval()

            return model
            # docs_tag: end_setup_pytorch

        # docs_tag: begin_setup_tensorrt
        elif self.backend == "tensorrt":
            # For TensorRT, the process is the following:
            # We check if there already exists a TensorRT engine generated
            # previously. If not, we check if there exists an ONNX model generated
            # previously. If not, we will generate both of the one by one
            # and then use those.
            # The underlying pytorch model that we use in case of TensorRT
            # inference is the FCN model from torchvision. It is only used during
            # the conversion process and not during the inference.
            onnx_file_path = os.path.join(
                self.results_dir,
                "model.%d.%d.%d.onnx"
                % (self.batch_size, self.target_img_height, self.target_img_height),
            )
            trt_engine_file_path = os.path.join(
                self.results_dir,
                "model.%d.%d.%d.trtmodel"
                % (self.batch_size, self.target_img_height, self.target_img_height),
            )

            torch_model = segmentation_models.fcn_resnet101
            weights_info = segmentation_models.FCN_ResNet101_Weights

            weights = weights_info.DEFAULT
            self.class_to_idx_dict = {
                cls: idx for (idx, cls) in enumerate(weights.meta["categories"])
            }

            if self.visualization_class_name not in self.class_to_idx_dict:
                print(
                    "Requested segmentation class '%s' is not supported by the "
                    "FCN model." % self.visualization_class_name
                )
                print(
                    "All supported class names are: %s"
                    % ", ".join(self.class_to_idx_dict.keys())
                )
                exit(1)

            # Check if we have a previously generated model.
            if not os.path.isfile(trt_engine_file_path):
                if not os.path.isfile(onnx_file_path):
                    # First we use PyTorch to create a segmentation model.
                    with torch.no_grad():
                        pyt_model = torch_model(weights=weights)
                        pyt_model.to("cuda")
                        pyt_model.eval()

                        # Allocate a dummy input to help generate an ONNX model.
                        dummy_x_in = torch.randn(
                            self.batch_size,
                            3,
                            self.target_img_height,
                            self.target_img_width,
                            requires_grad=False,
                        ).cuda()

                        # Generate an ONNX model using the PyTorch's onnx export.
                        torch.onnx.export(
                            pyt_model,
                            args=dummy_x_in,
                            f=onnx_file_path,
                            export_params=True,
                            opset_version=11,
                            do_constant_folding=True,
                            input_names=["input"],
                            output_names=["output"],
                            dynamic_axes={
                                "input": {0: "batch_size"},
                                "output": {0: "batch_size"},
                            },
                        )

                        # Remove the tensors and model after this.
                        del pyt_model
                        del dummy_x_in
                        torch.cuda.empty_cache()

                    print("Generated an ONNX model and saved at: %s" % onnx_file_path)
                else:
                    print("Using a pre-built ONNX model from: %s" % onnx_file_path)

                # Now that we have an ONNX model, we will continue generating a
                # serialized TensorRT engine from it.
                success = convert_onnx_to_tensorrt(
                    onnx_file_path,
                    trt_engine_file_path,
                    max_batch_size=self.batch_size,
                    max_workspace_size=1,
                )
                if success:
                    print("Generated TensorRT engine in: %s" % trt_engine_file_path)
                else:
                    print("Failed to generate the TensorRT engine.")
                    exit(1)

            else:
                print(
                    "Using a pre-built TensorRT engine from: %s" % trt_engine_file_path
                )

            # docs_tag: begin_load_tensorrt
            # Once the TensorRT engine generation is all done, we load it.
            trt_logger = trt.Logger(trt.Logger.INFO)
            with open(trt_engine_file_path, "rb") as f, trt.Runtime(
                trt_logger
            ) as runtime:
                trt_model = runtime.deserialize_cuda_engine(f.read())

            # Create execution context.
            context = trt_model.create_execution_context()

            # Allocate the output bindings.
            output_tensors, output_idx = setup_tensort_bindings(
                trt_model, self.device_id
            )

            return context, output_tensors, output_idx
            # docs_tag: end_setup_tensorrt

        else:
            print(
                "Invalid backend option specified: %s. "
                "Currently supports: pytorch, tensorrt" % self.backend
            )
            exit(1)

    # docs_tag: begin_execute_inference
    def execute_inference(self, model_info, torch_preprocessed_tensor):
        # Executes inference depending on the type of the backend.
        # docs_tag: begin_infer_pytorch
        if self.backend == "pytorch":
            with torch.no_grad():
                infer_output = model_info(torch_preprocessed_tensor)["out"]

            return infer_output
            # docs_tag: end_infer_pytorch

        # docs_tag: begin_infer_tensorrt
        elif self.backend == "tensorrt":
            # Setup TensorRT IO binding pointers.
            # docs_tag: begin_tensorrt_unpack
            context, output_tensors, output_idx = model_info  # Un-pack this.
            # docs_tag: end_tensorrt_unpack

            # We need to check the allocated batch size and the required batch
            # size. Sometimes, during to batching, the last batch may be of
            # less size than the batch size. In those cases, we would simply
            # pad that with zero inputs and discard its output later on.
            # docs_tag: begin_check_last_batch
            allocated_batch_size = output_tensors[output_idx].shape[0]
            required_batch_size = torch_preprocessed_tensor.shape[0]

            if allocated_batch_size != required_batch_size:
                # Need to pad the input with extra zeros tensors.
                new_input_shape = [allocated_batch_size - required_batch_size] + list(
                    torch_preprocessed_tensor.shape[1:]
                )

                # Allocate just the extra input required.
                extra_input = torch.zeros(
                    size=new_input_shape,
                    dtype=torch_preprocessed_tensor.dtype,
                    device=self.device_id,
                )

                # Update the existing input tensor by joining it with the newly
                # created input.
                torch_preprocessed_tensor = torch.cat(
                    (torch_preprocessed_tensor, extra_input)
                )
                # docs_tag: end_check_last_batch

            # docs_tag: begin_tensorrt_run
            # Prepare the TensorRT I/O bindings.
            input_bindings = [torch_preprocessed_tensor.data_ptr()]
            output_bindings = []
            for t in output_tensors:
                output_bindings.append(t.data_ptr())
            io_bindings = input_bindings + output_bindings

            # Execute synchronously.
            context.execute_v2(bindings=io_bindings)
            infer_output = output_tensors[output_idx]
            # docs_tag: end_tensorrt_run

            # docs_tag: begin_discard_last_batch
            # Finally, check if we had padded the input. If so, we need to
            # discard the extra output.
            if allocated_batch_size != required_batch_size:
                # We need remove the padded output.
                infer_output = torch.split(
                    infer_output,
                    [required_batch_size, allocated_batch_size - required_batch_size],
                )[0]

            return infer_output
            # docs_tag: end_discard_last_batch
            # docs_tag: end_infer_tensorrt

        else:
            print(
                "Invalid backend option specified: %s. "
                "Currently supports: pytorch, tensorrt" % self.backend
            )
            exit(1)

    # docs_tag: begin_run
    def run(self):
        # docs_tag: begin_run_basics
        # Runs the complete sample end-to-end.
        max_image_size = 1024 * 1024 * 3  # Maximum possible image size.

        # First setup the model.
        model_info = self.setup_model()

        # Next, we would batchify the file_list and data_list based on the
        # batch size and start processing these batches one by one.
        file_name_batches = [
            self.file_names[i : i + self.batch_size]  # noqa: E203
            for i in range(0, len(self.file_names), self.batch_size)
        ]
        data_batches = [
            self.data[i : i + self.batch_size]  # noqa: E203
            for i in range(0, len(self.data), self.batch_size)
        ]
        batch_idx = 0
        # docs_tag: end_run_basics

        # docs_tag: begin_batch_loop
        # We will use the torchnvjpeg based decoder on the GPU. This will be
        # allocated once during the first run or whenever a batch size change
        # happens.
        decoder = None

        for file_name_batch, data_batch in zip(file_name_batches, data_batches):
            print("Processing batch %d of %d" % (batch_idx + 1, len(file_name_batches)))
            effective_batch_size = len(file_name_batch)

            # docs_tag: begin_decode
            # Decode in batch using torchnvjpeg decoder on the GPU.
            if not decoder or effective_batch_size != self.batch_size:
                decoder = torchnvjpeg.Decoder(
                    0,
                    0,
                    True,
                    self.device_id,
                    effective_batch_size,
                    8,  # this is max_cpu_threads parameter. Not used internally.
                    max_image_size,
                    torch.cuda.current_stream(self.device_id),
                )
            image_tensor_list = decoder.batch_decode(data_batch)

            # Convert the list of tensors to a tensor itself.
            image_tensors = torch.stack(image_tensor_list)

            # Also save an NCHW version of the image tensors.
            image_tensors_nchw = image_tensors.permute(0, 3, 1, 2)  # from NHWC to NCHW

            input_image_height, input_image_width = (
                image_tensors.shape[1],
                image_tensors.shape[2],
            )
            # docs_tag: end_decode

            # docs_tag: begin_torch_to_cvcuda
            # A torch tensor can be wrapped into a CVCUDA Object using the "as_tensor"
            # function in the specified layout. The datatype and dimensions are derived
            # directly from the torch tensor.
            cvcuda_input_tensor = cvcuda.as_tensor(image_tensors, "NHWC")
            # docs_tag: end_torch_to_cvcuda

            # docs_tag: begin_preproc
            # Start the pre-processing now. For segmentation, pre-processing includes
            # the following sequence of operations.
            # Resize -> DataType Convert(U8->F32) -> Normalize -> Interleaved to Planar

            # Resize to the input network dimensions.
            cvcuda_resized_tensor = cvcuda.resize(
                cvcuda_input_tensor,
                (
                    effective_batch_size,
                    self.target_img_height,
                    self.target_img_width,
                    3,
                ),
                cvcuda.Interp.LINEAR,
            )

            # Convert to the data type and range of values needed by the input layer
            # i.e uint8->float. The values are first scaled to the 0-1 range.
            cvcuda_float_tensor = cvcuda.convertto(
                cvcuda_resized_tensor, np.float32, scale=1 / 255
            )

            # Normalize using mean and std-dev
            mean_tensor = torch.Tensor([0.485, 0.456, 0.406])
            stddev_tensor = torch.Tensor([0.229, 0.224, 0.225])
            mean_tensor = mean_tensor.reshape(1, 1, 1, 3).cuda(self.device_id)
            stddev_tensor = stddev_tensor.reshape(1, 1, 1, 3).cuda(self.device_id)
            cvcuda_mean_tensor = cvcuda.as_tensor(mean_tensor, "NHWC")
            cvcuda_stddev_tensor = cvcuda.as_tensor(stddev_tensor, "NHWC")
            cvcuda_normalized_tensor = cvcuda.normalize(
                cvcuda_float_tensor,
                base=cvcuda_mean_tensor,
                scale=cvcuda_stddev_tensor,
                flags=cvcuda.NormalizeFlags.SCALE_IS_STDDEV,
            )

            # The final stage in the pre-process pipeline includes converting the NHWC
            # buffer into a NCHW buffer.
            cvcuda_preprocessed_tensor = cvcuda.reformat(
                cvcuda_normalized_tensor, "NCHW"
            )
            # docs_tag: end_preproc

            # docs_tag: begin_run_infer
            # Execute the inference after converting the tensor back to Torch.
            torch_preprocessed_tensor = torch.as_tensor(
                cvcuda_preprocessed_tensor.cuda(),
                device=torch.device("cuda", self.device_id),
            )
            infer_output = self.execute_inference(model_info, torch_preprocessed_tensor)
            # docs_tag: end_run_infer

            # Once the inference is over we would start the post-processing steps.
            # First, we normalize the probability scores from the network.
            # docs_tag: begin_normalize
            normalized_masks = torch.nn.functional.softmax(infer_output, dim=1)
            # docs_tag: end_normalize

            # Then filter based on the scores corresponding only to the class of
            # interest
            # docs_tag: begin_argmax
            class_masks = (
                normalized_masks.argmax(dim=1)
                == self.class_to_idx_dict[self.visualization_class_name]
            )
            class_masks = torch.unsqueeze(class_masks, dim=-1)  # Makes it NHWC
            class_masks = class_masks.type(torch.uint8)  # Make it uint8 from bool
            # docs_tag: end_argmax

            # Then convert the masks back to CV-CUDA tensor for rest of the
            # post-processing:
            # 1) Up-scaling back to the original image dimensions
            # 2) Apply blur on the original images and overlay on the original image.

            # Convert back to CV-CUDA tensor
            # docs_tag: begin_mask_upscale
            cvcuda_class_masks = cvcuda.as_tensor(class_masks.cuda(), "NHWC")
            # Upscale it.
            cvcuda_class_masks_upscaled = cvcuda.resize(
                cvcuda_class_masks,
                (effective_batch_size, input_image_height, input_image_width, 1),
                cvcuda.Interp.LINEAR,
            )
            # Convert back to PyTorch.
            class_masks_upscaled = torch.as_tensor(
                cvcuda_class_masks_upscaled.cuda(),
                device=torch.device("cuda", self.device_id),
            )
            # Repeat in last dimension to make the mask 3 channel
            class_masks_upscaled = class_masks_upscaled.repeat(1, 1, 1, 3)
            class_masks_upscaled_nchw = class_masks_upscaled.permute(
                0, 3, 1, 2
            )  # from NHWC to NCHW
            # docs_tag: end_mask_upscale

            # Blur the input images using the median blur op and convert to PyTorch.
            # docs_tag: begin_input_blur
            cvcuda_blurred_input_imgs = cvcuda.median_blur(
                cvcuda_input_tensor, ksize=(27, 27)
            )
            cvcuda_blurred_input_imgs = cvcuda.reformat(
                cvcuda_blurred_input_imgs, "NCHW"
            )
            blurred_input_imgs = torch.as_tensor(
                cvcuda_blurred_input_imgs.cuda(),
                device=torch.device("cuda", self.device_id),
            )
            # docs_tag: end_input_blur

            # Create an overlay image. We do this by selectively blurring out pixels
            # in the input image where the class mask prediction was absent (i.e. False)
            # We already have all the things required for this: The input images,
            # the blurred version of the input images and the upscale version
            # of the mask
            # docs_tag: begin_overlay
            mask_absent = class_masks_upscaled_nchw == 0
            image_tensors_nchw[mask_absent] = blurred_input_imgs[
                mask_absent
            ]  # In-place
            # docs_tag: end_overlay

            # Loop over all the images in the current batch and save the
            # inference results.
            # docs_tag: begin_visualization_loop
            for img_idx in range(effective_batch_size):
                img_name = os.path.splitext(os.path.basename(file_name_batch[img_idx]))[
                    0
                ]
                results_path = os.path.join(self.results_dir, "out_%s.jpg" % img_name)
                print(
                    "\tSaving the overlay result for %s class for to: %s"
                    % (self.visualization_class_name, results_path)
                )

                # Convert the overlay which was in-place saved in
                # image_tensors_nchw to a PIL image on the CPU and save it.
                overlay_cpu = image_tensors_nchw[img_idx].detach().cpu()
                overlay_pil = F.to_pil_image(overlay_cpu)
                overlay_pil.save(results_path)

            # Increment the batch counter.
            batch_idx += 1
            # docs_tag: end_visualization_loop


# docs_tag: main_func
def main():
    parser = argparse.ArgumentParser(
        "Semantic segmentation sample using CV-CUDA.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "-i",
        "--input_path",
        default="./assets/Weimaraner.jpg",
        type=str,
        help="Either a path to a JPEG image or a directory containing JPEG "
        "images to use as input.",
    )

    parser.add_argument(
        "-o",
        "--output_dir",
        default="/tmp",
        type=str,
        help="The folder where the output segmentation overlay should be stored.",
    )

    parser.add_argument(
        "-c",
        "--class_name",
        default="dog",
        type=str,
        help="The segmentation class to visualize the results for.",
    )

    parser.add_argument(
        "-th",
        "--target_img_height",
        default=224,
        type=int,
        help="The height to which you want to resize the input_image before "
        "running inference.",
    )

    parser.add_argument(
        "-tw",
        "--target_img_width",
        default=224,
        type=int,
        help="The width to which you want to resize the input_image before "
        "running inference.",
    )

    parser.add_argument(
        "-b",
        "--batch_size",
        default=1,
        type=int,
        help="Artificially simulated batch size. The same input image will be read and "
        "used this many times. Useful for performance bench-marking.",
    )

    parser.add_argument(
        "-d",
        "--device_id",
        default=0,
        type=int,
        help="The GPU device to use for this sample.",
    )

    parser.add_argument(
        "-bk",
        "--backend",
        default="pytorch",
        type=str,
        help="The inference backend to use. Currently supports pytorch, tensorrt.",
    )

    # Parse the command line arguments.
    args = parser.parse_args()

    # Run the sample.
    sample = SemanticSegmentationSample(
        args.input_path,
        args.output_dir,
        args.class_name,
        args.batch_size,
        args.target_img_height,
        args.target_img_width,
        args.device_id,
        args.backend,
    )

    sample.run()


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cvcuda
import numpy as np
import util


RNG = np.random.default_rng(0)


def test_op_channelreorder_varshape():

    input = util.create_image_batch(10, cvcuda.Format.RGB8, size=(123, 321), rng=RNG)
    order = util.create_tensor((10, 3), np.int32, "NC", max_random=(2, 2, 2), rng=RNG)

    out = cvcuda.channelreorder(input, order)
    assert len(out) == len(input)
    assert out.capacity == input.capacity
    assert out.uniqueformat == input.uniqueformat
    assert out.maxsize == input.maxsize

    order = util.create_tensor(
        (10, 4), np.int32, "NC", max_random=(3, 3, 3, 3), rng=RNG
    )
    out = cvcuda.channelreorder(input, order, format=cvcuda.Format.BGRA8)

    assert len(out) == len(input)
    assert out.capacity == input.capacity
    assert out.uniqueformat == cvcuda.Format.BGRA8
    assert out.maxsize == input.maxsize

    stream = cvcuda.Stream()
    out = util.clone_image_batch(input)
    tmp = cvcuda.channelreorder_into(src=input, dst=out, orders=order, stream=stream)
    assert tmp is out
    assert len(out) == len(input)
    assert out.capacity == input.capacity
    assert out.uniqueformat == input.uniqueformat
    assert out.maxsize == input.maxsize



#!/usr/bin/env python
#
# Copyright 2008, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Verifies that Google Test correctly determines whether to use colors."""

import os
from googletest.test import gtest_test_utils

IS_WINDOWS = os.name == 'nt'

COLOR_ENV_VAR = 'GTEST_COLOR'
COLOR_FLAG = 'gtest_color'
COMMAND = gtest_test_utils.GetTestExecutablePath('googletest-color-test_')


def SetEnvVar(env_var, value):
  """Sets the env variable to 'value'; unsets it when 'value' is None."""

  if value is not None:
    os.environ[env_var] = value
  elif env_var in os.environ:
    del os.environ[env_var]


def UsesColor(term, color_env_var, color_flag):
  """Runs googletest-color-test_ and returns its exit code."""

  SetEnvVar('TERM', term)
  SetEnvVar(COLOR_ENV_VAR, color_env_var)

  if color_flag is None:
    args = []
  else:
    args = ['--%s=%s' % (COLOR_FLAG, color_flag)]
  p = gtest_test_utils.Subprocess([COMMAND] + args)
  return not p.exited or p.exit_code


class GTestColorTest(gtest_test_utils.TestCase):
  def testNoEnvVarNoFlag(self):
    """Tests the case when there's neither GTEST_COLOR nor --gtest_color."""

    if not IS_WINDOWS:
      self.assert_(not UsesColor('dumb', None, None))
      self.assert_(not UsesColor('emacs', None, None))
      self.assert_(not UsesColor('xterm-mono', None, None))
      self.assert_(not UsesColor('unknown', None, None))
      self.assert_(not UsesColor(None, None, None))
    self.assert_(UsesColor('linux', None, None))
    self.assert_(UsesColor('cygwin', None, None))
    self.assert_(UsesColor('xterm', None, None))
    self.assert_(UsesColor('xterm-color', None, None))
    self.assert_(UsesColor('xterm-kitty', None, None))
    self.assert_(UsesColor('xterm-256color', None, None))

  def testFlagOnly(self):
    """Tests the case when there's --gtest_color but not GTEST_COLOR."""

    self.assert_(not UsesColor('dumb', None, 'no'))
    self.assert_(not UsesColor('xterm-color', None, 'no'))
    if not IS_WINDOWS:
      self.assert_(not UsesColor('emacs', None, 'auto'))
    self.assert_(UsesColor('xterm', None, 'auto'))
    self.assert_(UsesColor('dumb', None, 'yes'))
    self.assert_(UsesColor('xterm', None, 'yes'))

  def testEnvVarOnly(self):
    """Tests the case when there's GTEST_COLOR but not --gtest_color."""

    self.assert_(not UsesColor('dumb', 'no', None))
    self.assert_(not UsesColor('xterm-color', 'no', None))
    if not IS_WINDOWS:
      self.assert_(not UsesColor('dumb', 'auto', None))
    self.assert_(UsesColor('xterm-color', 'auto', None))
    self.assert_(UsesColor('dumb', 'yes', None))
    self.assert_(UsesColor('xterm-color', 'yes', None))

  def testEnvVarAndFlag(self):
    """Tests the case when there are both GTEST_COLOR and --gtest_color."""

    self.assert_(not UsesColor('xterm-color', 'no', 'no'))
    self.assert_(UsesColor('dumb', 'no', 'yes'))
    self.assert_(UsesColor('xterm-color', 'no', 'auto'))

  def testAliasesOfYesAndNo(self):
    """Tests using aliases in specifying --gtest_color."""

    self.assert_(UsesColor('dumb', None, 'true'))
    self.assert_(UsesColor('dumb', None, 'YES'))
    self.assert_(UsesColor('dumb', None, 'T'))
    self.assert_(UsesColor('dumb', None, '1'))

    self.assert_(not UsesColor('xterm', None, 'f'))
    self.assert_(not UsesColor('xterm', None, 'false'))
    self.assert_(not UsesColor('xterm', None, '0'))
    self.assert_(not UsesColor('xterm', None, 'unknown'))


if __name__ == '__main__':
  gtest_test_utils.Main()



# SPDX-FileCopyrightText: Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cvcuda
import pytest as t
from random import randint


@t.mark.parametrize(
    "num_images, format, min_out_size, max_out_size, border_mode, border_value",
    [
        (
            4,
            cvcuda.RGBf32,
            (1, 1),
            (128, 128),
            cvcuda.Border.CONSTANT,
            [0],
        ),
        (
            5,
            cvcuda.RGB8,
            (1, 1),
            (128, 128),
            cvcuda.Border.CONSTANT,
            [12, 3, 4, 55],
        ),
        (
            9,
            cvcuda.RGBA8,
            (1, 1),
            (128, 128),
            cvcuda.Border.WRAP,
            [0],
        ),
        (
            12,
            cvcuda.RGBAf32,
            (1, 1),
            (128, 128),
            cvcuda.Border.REPLICATE,
            [0],
        ),
        (
            8,
            cvcuda.RGB8,
            (1, 1),
            (128, 128),
            cvcuda.Border.REFLECT,
            [0],
        ),
        (
            10,
            cvcuda.RGBA8,
            (1, 1),
            (128, 128),
            cvcuda.Border.REFLECT101,
            [0],
        ),
    ],
)
def test_op_copymakeborder(
    num_images, format, min_out_size, max_out_size, border_mode, border_value
):
    max_out_w = randint(min_out_size[0], max_out_size[0])
    max_out_h = randint(min_out_size[1], max_out_size[1])

    input = cvcuda.ImageBatchVarShape(num_images)
    varshape_out = cvcuda.ImageBatchVarShape(num_images)
    out_heights = []
    out_widths = []
    for i in range(num_images):
        w = randint(1, max_out_w)
        h = randint(1, max_out_h)
        img_i = cvcuda.Image([w, h], format)
        input.pushback(img_i)
        w_out = randint(w, max_out_size[0])
        h_out = randint(h, max_out_size[1])
        img_o = cvcuda.Image([w_out, h_out], format)
        varshape_out.pushback(img_o)
        out_heights.append(h_out)
        out_widths.append(w_out)

    top_tensor = cvcuda.Tensor([1, 1, num_images, 1], cvcuda.Type.S32, "NHWC")
    left_tensor = cvcuda.Tensor([1, 1, num_images, 1], cvcuda.Type.S32, "NHWC")

    tensor_out = cvcuda.Tensor(num_images, [max_out_w, max_out_h], format)

    out = cvcuda.copymakeborderstack(
        input,
        top=top_tensor,
        left=left_tensor,
        out_height=max_out_h,
        out_width=max_out_w,
    )
    assert out.layout == tensor_out.layout
    assert out.shape == tensor_out.shape
    assert out.dtype == tensor_out.dtype

    stream = cvcuda.Stream()
    tmp = cvcuda.copymakeborderstack_into(
        src=input,
        dst=tensor_out,
        top=top_tensor,
        left=left_tensor,
        border_mode=border_mode,
        border_value=border_value,
        stream=stream,
    )
    assert tmp is tensor_out

    out = cvcuda.copymakeborder(
        src=input,
        top=top_tensor,
        left=left_tensor,
        out_heights=out_heights,
        out_widths=out_widths,
        stream=stream,
    )
    assert out.uniqueformat is not None
    assert out.uniqueformat == varshape_out.uniqueformat
    for res, ref in zip(out, varshape_out):
        assert res.size == ref.size
        assert res.format == ref.format

    tmp = cvcuda.copymakeborder_into(
        src=input,
        dst=varshape_out,
        top=top_tensor,
        left=left_tensor,
        border_mode=border_mode,
        border_value=border_value,
        stream=stream,
    )
    assert tmp is varshape_out



# SPDX-FileCopyrightText: Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import nvcv
import numpy as np
import numbers
import torch
import copy


DTYPE = {
    nvcv.Format.HSV8: np.uint8,
    nvcv.Format.BGRA8: np.uint8,
    nvcv.Format.RGBA8: np.uint8,
    nvcv.Format.BGR8: np.uint8,
    nvcv.Format.RGB8: np.uint8,
    nvcv.Format.RGBA8: np.uint8,
    nvcv.Format.RGBf32: np.float32,
    nvcv.Format.F32: np.float32,
    nvcv.Format.U8: np.uint8,
    nvcv.Format.Y8: np.uint8,
    nvcv.Format.Y8_ER: np.uint8,
    nvcv.Format.U16: np.uint16,
    nvcv.Format.S16: np.int16,
    nvcv.Format.S32: np.int32,
}


def dist_odd(x):
    """Add one to x if even to make it odd

    Args:
        x (number): Value to turn odd if even

    Returns:
        number: Odd input value
    """
    return x if x % 2 == 1 else x + 1


def generate_data(shape, dtype, max_random=None, rng=None):
    """Generate data as numpy array

    Args:
        shape (tuple or list): Data shape
        dtype (numpy dtype): Data type (e.g. np.uint8)
        max_random (number or tuple or list): Maximum random value
        rng (numpy random Generator): To fill data with random values

    Returns:
        numpy.array: The generated data
    """
    if rng is None:
        data = np.zeros(shape, dtype=dtype)
    else:
        if max_random is not None and type(max_random) in {tuple, list}:
            assert len(max_random) == shape[-1]
        if issubclass(dtype, numbers.Integral):
            if max_random is None:
                max_random = [np.iinfo(dtype).max for _ in range(len(shape))]
            data = rng.integers(max_random, size=shape, dtype=dtype)
        elif issubclass(dtype, numbers.Real):
            if max_random is None:
                max_random = [1.0 for _ in range(len(shape))]
            data = rng.random(size=shape, dtype=dtype) * np.array(max_random)
            data = data.astype(dtype)
    return data


def to_cuda_buffer(host):
    orig_dtype = copy.copy(host.dtype)

    # torch doesn't accept uint16. Let's make it believe
    # it is handling int16 instead.
    if host.dtype == np.uint16:
        host.dtype = np.int16

    dev = torch.as_tensor(host, device="cuda").cuda()
    host.dtype = orig_dtype  # restore it

    class CudaBuffer:
        __cuda_array_interface = None
        obj = None

    # The cuda buffer only needs the cuda array interface.
    # We can then set its dtype to whatever we want.
    buf = CudaBuffer()
    buf.__cuda_array_interface__ = dev.__cuda_array_interface__
    buf.__cuda_array_interface__["typestr"] = orig_dtype.str
    buf.obj = dev  # make sure it holds a reference to the torch buffer

    return buf


def create_tensor(shape, dtype, layout, max_random=None, rng=None, transform_dist=None):
    """Create a tensor

    Args:
        shape (tuple or list): Tensor shape
        dtype (numpy dtype): Tensor data type (e.g. np.uint8)
        layout (string): Tensor layout (e.g. NC, HWC, NHWC)
        max_random (number or tuple or list): Maximum random value
        rng (numpy random Generator): To fill tensor with random values
        transform_dist (function): To transform random values (e.g. MAKE_ODD)

    Returns:
        nvcv.Tensor: The created tensor
    """
    h_data = generate_data(shape, dtype, max_random, rng)
    if transform_dist is not None:
        vec_transform_dist = np.vectorize(transform_dist)
        h_data = vec_transform_dist(h_data)
        h_data = h_data.astype(dtype)
    tensor = nvcv.as_tensor(to_cuda_buffer(h_data), layout=layout)
    return tensor


def create_image(size, img_format, max_random=None, rng=None):
    """Create an image

    Args:
        size (tuple or list): Image size (width, height)
        img_format (nvcv.Format): Image format
        max_random (number or tuple or list): Maximum random value
        rng (numpy random Generator): To image with random values

    Returns:
        nvcv.Image: The created image
    """
    shape = (size[1], size[0], img_format.channels)
    dtype = DTYPE[img_format]
    h_data = generate_data(shape, dtype, max_random, rng)
    image = nvcv.as_image(to_cuda_buffer(h_data))
    return image


def create_image_batch(
    num_images, img_format, size=(0, 0), max_size=(128, 128), max_random=None, rng=None
):
    """Create an image batch

    Args:
        num_images (number): Number of images in the batch
        img_format (nvcv.ImageFormat): Image format of each image
        size (tuple or list): Image size (width, height) use (0, 0) for random sizes
        max_size (tuple or list): Use random image size from 1 to max_size
        max_random (number or tuple or list): Maximum random value inside each image
        rng (numpy random Generator): To fill each image with random values

    Returns:
        nvcv.ImageBatchVarShape: The created image batch
    """
    if size[0] == 0 or size[1] == 0:
        assert rng is not None
    image_batch = nvcv.ImageBatchVarShape(num_images)
    for i in range(num_images):
        w = (
            rng.integers(1, max_size[0] + 1)
            if rng is not None and size[0] == 0
            else size[0]
        )
        h = (
            rng.integers(1, max_size[1] + 1)
            if rng is not None and size[1] == 0
            else size[1]
        )
        image_batch.pushback(create_image((w, h), img_format, max_random, rng))
    return image_batch


def clone_image_batch(input_image_batch, img_format=None):
    """Clone an image batch

    Args:
        input_image_batch (nvcv.ImageBatchVarShape): Image batch to be cloned
        img_format (nvcv.ImageFormat): Image format of the output

    Returns:
        nvcv.ImageBatchVarShape: The cloned image batch var shape
    """
    output_image_batch = nvcv.ImageBatchVarShape(input_image_batch.capacity)
    for input_image in input_image_batch:
        image = nvcv.Image(
            input_image.size, input_image.format if img_format is None else img_format
        )
        output_image_batch.pushback(image)
    return output_image_batch



# SPDX-FileCopyrightText: Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
trt_utils

This file hosts various TensorRT related utilities.
"""

import torch
import numpy as np
import tensorrt as trt


def convert_onnx_to_tensorrt(
    onnx_file_path, trt_engine_file_path, max_batch_size, max_workspace_size=5
):
    """
    Converts an ONNX engine to a serialized TensorRT engine.
    :param onnx_file_path: Full path to an existing ONNX file.
    :param trt_engine_file_path: Full path to save the generated TensorRT Engine file.
    :param max_batch_size: The maximum batch size to use in the TensorRT engine.
    :param max_workspace_size: The maximum GPU memory that TensorRT can use (in GB.)
    :return: True if engine was generated. False otherwise.
    """
    print("Using TensorRT version: %s" % trt.__version__)
    trt_logger = trt.Logger(trt.Logger.INFO)

    with trt.Builder(trt_logger) as builder, builder.create_network(
        1 << int(trt.NetworkDefinitionCreationFlag.EXPLICIT_BATCH)
    ) as network, trt.OnnxParser(network, trt_logger) as parser:

        config = builder.create_builder_config()
        config.set_memory_pool_limit(
            trt.MemoryPoolType.WORKSPACE, 1024 * 1024 * 1024 * max_workspace_size
        )  # Sets workspace size in GB.
        if builder.platform_has_fast_fp16:
            config.set_flag(trt.BuilderFlag.FP16)
            config.set_flag(trt.BuilderFlag.STRICT_TYPES)
            print("Using precision : float16")
        else:
            print("Using precision : float32")

        # Parse model file
        print("Loading ONNX file from path %s " % onnx_file_path)
        with open(onnx_file_path, "rb") as model:
            if not parser.parse(model.read()):
                print("Failed to parse the ONNX file.")
                for error in range(parser.num_errors):
                    print(parser.get_error(error))
                return False

        for input_idx in range(network.num_inputs):
            new_shape = [max_batch_size]
            new_shape.extend(network.get_input(input_idx).shape[1:])
            network.get_input(input_idx).shape = new_shape

        for input_idx in range(network.num_inputs):
            print(
                "INPUT[%d] : %s %s"
                % (
                    input_idx,
                    network.get_input(input_idx).name,
                    network.get_input(input_idx).shape,
                )
            )
        for output_idx in range(network.num_outputs):
            print(
                "OUTPUT[%d] : %s %s"
                % (
                    output_idx,
                    network.get_output(output_idx).name,
                    network.get_output(output_idx).shape,
                )
            )
        print("Completed parsing of ONNX file.")

        print(
            "Building an engine from file %s. This may take a while..." % onnx_file_path
        )

        engine = builder.build_serialized_network(network, config)

        if engine:
            print("Completed creating Engine. Saving on disk...")
            with open(trt_engine_file_path, "wb") as f:
                f.write(engine)
                print("Saved to file %s" % trt_engine_file_path)
            return True
        else:
            print("Failed in creating the TensorRT engine.")
            return False


def setup_tensort_bindings(trt_model, device_id):
    """
    Setups the I/O bindings for a TensorRT engine for the first time.
    :param trt_model: Full path to the generated TensorRT Engine file.
    :param device_id: The GPU device id on which you want to allocated the buffers.
    :return: A list of output tensors and the index of the first output.
    """
    # For TensorRT, we need to allocate the output data buffers.
    # The input data buffers are already allocated by us.
    output_binding_idx = 0
    output_idx = 0
    output_tensors = []

    # Loop over all the I/O bindings.
    for b_idx in range(trt_model.num_io_tensors):
        # Get various properties associated with the bindings.
        b_name = trt_model.get_tensor_name(b_idx)
        b_shape = tuple(trt_model.get_tensor_shape(b_name))
        b_dtype = np.dtype(trt.nptype(trt_model.get_tensor_dtype(b_name))).name

        print(
            "TensorRT Binding[%d]: %s := shape: %s dtype: %s"
            % (b_idx, b_name, str(b_shape), b_dtype)
        )

        # Append to the appropriate list.
        if trt_model.get_tensor_mode(b_name) == trt.TensorIOMode.OUTPUT:
            # First allocate on device output buffers, using PyTorch.
            output = torch.zeros(
                size=b_shape, dtype=getattr(torch, b_dtype), device=device_id
            )

            print("\tAllocated the binding as an output.")
            # Since we know the name of our output layer, we will check against
            # it and grab its binding index.
            if b_name == "output":
                output_idx = output_binding_idx

            output_binding_idx += 1
            output_tensors.append(output)

    return output_tensors, output_idx



import pytest

asyncio = pytest.importorskip("asyncio")
m = pytest.importorskip("pybind11_tests.async_module")


@pytest.fixture
def event_loop():
    loop = asyncio.new_event_loop()
    yield loop
    loop.close()


async def get_await_result(x):
    return await x


def test_await(event_loop):
    assert 5 == event_loop.run_until_complete(get_await_result(m.SupportsAsync()))


def test_await_missing(event_loop):
    with pytest.raises(TypeError):
        event_loop.run_until_complete(get_await_result(m.DoesNotSupportAsync()))
